/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "common.h"
#include "r5_sys_mutex.h"
#include "r5_multiboot.h"
#include "xpseudo_asm_gcc.h"

/*
 * NOTE: This code runs on r5_0. Another app may run on r5_1.
 * Thus, the common registers for EMIO MUST be initialized
 * in the reginit section of the PDI, so that either app can
 * run first.
 */

void r5InitMultibootState(MultibootState* pMS)
{
    uint32_t qTmp;

    /*
     * Inherits clocks/resets from PLM initialization
     */
    r5GrabSysMutex();
    {
        /*
         * MASK_DATA_3_LSW: R5_UP is only output, assert its mask and value
         * The MSW is a separate register allocated to other R5
         */
        *(volatile uint32_t*)0xFF0B0018 = (~EMIO_R5_UP << 16) | (~EMIO_R5_UP & 0xffff);

        /*
         * DIRM_3: R5_UP is output.
         * This is a shared register, do read-modify-write
         */
        qTmp = *(volatile uint32_t*)0xFF0B02C4;
        qTmp |= EMIO_R5_UP;
        *(volatile uint32_t*)0xFF0B02C4 = qTmp;

        /*
         * OEN_3: R5_UP is enabled.
         * This is a shared register, do read-modify-write
         */
        qTmp = *(volatile uint32_t*)0xFF0B02C8;
        qTmp |= EMIO_R5_UP;
        *(volatile uint32_t*)0xFF0B02C8 = qTmp;
    }
    r5ReleaseSysMutex();

    /*
     * Store next multiboot value
     */
#ifdef MULTIBOOT_VAL
    pMS->qMultibootReg = (uint32_t)MULTIBOOT_VAL;
#else
    pMS->qMultibootReg = *(volatile uint32_t*)0xF1110004 + 1;
#endif

    pMS->qPcieGoodCounter = EMIO_PCIE_GOOD_COUNTER_TIMEOUT;
    pMS->ePcieState = PCIE_STATE_UNKNOWN;
}

void r5SenseMultibootInputs(MultibootState* pMS)
{
    uint32_t qTmp;

    /*
     * Read EMIOs
     */
    qTmp = *(volatile uint32_t*)0xFF0B006C;

    /*
     * Both MULTIBOOT and CPLD_UP have to be asserted
     */
    pMS->yMultibootRequest = (0x0 == (qTmp & (EMIO_MULTIBOOT | EMIO_CPLD_UP)));
    if(qTmp & EMIO_PCIE_GOOD) {
        pMS->ePcieState = PCIE_STATE_UP;
        /*
         * Allows for flicker.
         * The bit will toggle once as it trains
         * The r5_app may or may not catch it
         */
        if(pMS->qPcieGoodCounter > 0) {
            pMS->qPcieGoodCounter--;
        }
    } else if(pMS->qPcieGoodCounter > 0) {
        pMS->qPcieGoodCounter--;
        pMS->ePcieState = PCIE_STATE_UNKNOWN;
    } else {
        pMS->ePcieState = PCIE_STATE_DOWN;
    }
}

void r5PerformMultiboot(MultibootState* pMS)
{
    uint32_t qTmp;

    TRACE();
    /*
     * Store scaled multiboot value into PMC_MULTI_BOOT
     */
    *(volatile uint32_t*)0xF1110004 = pMS->qMultibootReg;

    /*
     * Set boot mode user to qspi32
     */
    *(volatile uint32_t*)0xF1260200 = (2 << 12) | (1 << 8);

    if(0x20 == (*(volatile uint32_t*)0xF11A0004 & 0xff)) {
        TRACE();
        /*
         * Production PMC
         */
    } else {
        TRACE();

        /*
         * ES PMC
         */

        /*
         * Set SYSMON_REF_CTRL::SrcSel to PMC_IRO_CLK /2
         */
        *(volatile uint32_t*)0xF1260138 = 0;

        /*
         * Set RST_NONPS, all resets
         */
        *(volatile uint32_t*)0xF1260320 = 0x77;
    }
    TRACE();

    /*
     * Set PMC_GLOBAL::DOMAIN_ISO_CTRL::PMC_PL ("Isolation gasket")
     */
    qTmp = *(volatile uint32_t*)0xF1120000;
    qTmp |= 0x1000;
    *(volatile uint32_t*)0xF1120000 = qTmp;

    /*
     * Ensure all activity has finished before performing SRST
     */
    dsb();
    isb();

    /*
     * Perform SRST
     *
     * Write to RST_PS[PMC_SRST]
     */
    *(volatile uint32_t*)0xF126031C = 1 << 3;

    __builtin_unreachable();

    /*
     * Not reached
     */
    for(;;) {
        TRACE();
    }
}
