/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*****************************************************************************/
/**
* @file  r5_rem_app.c
*
* Example of how to do runtime error management (REM) and alert the PL via EMIO
* output.
* 
* There are 3 classes of messages driving respective EMIO pins:
*  -  "Config Done"   -- Level signal memory controller setup completed error free
*  -  "ECC Correction"-- Pulsed signal indicating a single-bit error was corrected
*  -  "Fatal Error"   -- Level signal indicating serious error detected at 
*                          start-up or during runtime.  
*
* @note
* This is intended to be an example that should be extended to cover 
* specific application needs. 
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver  Who    Date     Changes
* ---- ------ -------- ---------------------------------------------
* 1.00 jt     09/13/21 First release.  Interupt setup modelled 
*                      after xttcps_intr_example.c in the embeddedsw repo.
*</pre>
******************************************************************************/

/***************************** Include Files *********************************/

// Enable debug prints to console
#define DO_TRACE
// Enable self-test interrupt triggers
//#define DO_TEST

#include "common.h"
#include "xparameters.h"
#include "xstatus.h"
#include "xil_exception.h"
#include "xscugic.h"
#include "r5_sys_mutex.h"

#ifdef DO_TRACE
#include "xil_io.h"
#endif

/************************** Constant Definitions *****************************/

/*
 * Define how many Memory controllers the device has (regardless of usage)
 *
 * Versal 1902 device has 4 DDR controllers
 */
#define REM_NUM_CONTROLLERS 4

/*
 * Define the bitmap of ENABLED controllers
 *
 * Bit[0] maps to memory controller 0.
 * ...
 * Bit[3] maps to memory controller 3.
 *
 * Configure the bitmap as needed for your application.
 */
#define REM_ENABLED_MC_MAP   0x1 

/*
 * Define the bitmap of which memory controllers are using both memory channels
 *
 * Each memory controller may operate in single or dual-channel mode depending on the 
 * type of memory devices attached.  Some of the error signals are reported on a per-
 * channel basis, and we need to ignore the second channel when it's not used.  
 *
 * Bit[0] should be set when mem controller 0 is in dual-channel mode
 * ...
 * Bit[3] should be set when mem controller 3 is in dual-channel mode
 *
 * If the mem controller is not active via REM_ENABLED_MC_MAP, the respective bit here 
 * is a "don't care"
 */ 
#define REM_DUAL_CHANNEL_MAP  0x1 

/*
 * Define the masks for writing the respective EMIO bit for each status class.
 * 
 * The REM_MASK_DATA_3_MSW register manipulates EMIO bits 31:16.  
 */
#define REM_GPIO_CFGDONE_MASK 0x1 /* Assign EMIO bit[16] to 'config done'          */
#define REM_GPIO_CECC_MASK    0x2 /* Assign EMIO bit[17] to 'correctable ECC error */
#define REM_GPIO_FATAL_MASK   0x4 /* Assign EMIO bit[18] to 'Fatal Error'          */

// EMIO LPD_GPIO Bank 3 Configuration registers
#define REM_MASK_DATA_3_MSW 0xFF0B001C  /* EMIO Bank 3 upper 16-bits output reg     */
#define REM_DIRM_3          0xFF0B02C4  /* EMIO Bank 3 Direction mode 1=output      */
#define REM_OEN_3           0xFF0B02C8  /* EMIO Bank 3 output enable                */



#define REM_UB_PCSR_STATUS_OFFSET 0x8 /* Start-up status register offset  */

/*
 * Key value to unlock registers with write protection. Safest to assume 
 * the unlock event is good for only a single protected transaction. This code
 * manually re-locks upon completely the write.
 */
#define REM_PCSR_KEY 0xF9E8D7C6 


// Number if interrupts monitored for REM
#define REM_NUM_INTS   1
#define REM_NUM_POLL   1

#define REM_DATA_TABLE_SZ REM_NUM_INTS + REM_NUM_POLL

/*
 * Maximum Wait for CALDONE Timeout Delay
 *   The appropriate time is highly appication dependant on: 
 *     -- ECC fill is enabled or not
 *     -- memory speed 
 *     -- memory size
 *  Upper bound could be a minute wait in very extreme cases
 *  Timeout units: mS   --> 3000 means 3 seconds.
 */
#define REM_TIMEOUT_WAIT 3000 


#define REM_SUCCESS 0
#define REM_FAILURE 1

/**************************** Type Definitions *******************************/

/*
 * Per interrupt data table to store register base and offsets and masks.
 *  Intent is to be generic, but, each interrupt is constructed differently
 *  so this is the union of all needs.
 *
 * Status2MaskMap is a per-bit look-up table needed by the UB registers to map
 *  the fatal error status bits position to their mask/enable/disablle bits.
 *  The UB interrupts are contstructed very oddly with the mask/enable/disable
 *  using arbitrary bit-positions relative to the status/trigger registers.
 *  Use it as: The value at given array position (bit position in status register) 
 *             maps that status to the bit-number of the mask/enable/disable 
 *             register.
 */
typedef struct IntDataTable {
  u16 IRQ;
  u32 BaseAddr[REM_NUM_CONTROLLERS];
  u32 CEMaskCh0;  // Corretable ECC
  u32 CEMaskCh1;
  u32 FatalMaskCh0;
  u32 FatalMaskCh1;
  u32 LockOffset;
  u16 IsrOffset;  // status  -- wtc 
  u16 ItrOffset;  // trigger -- wo
  u16 ImrOffset;  // Mask rd -- ro
  u16 IerOffset;  // enable  -- wo
//u16 IdrOffset;  // disable -- wo   
} IntDataTable;

/***************** Macros (Inline Functions) Definitions *********************/


/************************** Function Prototypes ******************************/
static int  SetupInterruptSystem(u16 IntcDeviceID, XScuGic* IntcInstancePtr);
static void REMInterruptHandler(void* data);
static int  CheckDDRInitStatus(void);
static void wrGPIO(u16 enable, u16 data);
static void UnmaskMCBits(IntDataTable* RowPtr);


/************************** Variable Definitions *****************************/

static IntDataTable UBIntData = {
  .IRQ          = 178U,
  .BaseAddr     = { 0xF6110000, 0xF6280000, 0xF63F0000, 0xF6560000 }, 
                            // Base address of each controller
  .CEMaskCh0    = 0x0,      // Corretable ECC -- none
  .CEMaskCh1    = 0x0,      // Not Channelized 
  .FatalMaskCh0 = 0x1 << 2  // instr_ue     --> Status bit[2], mask bit[0]
                | 0x1 << 3  // data_ue      --> status bit[3], mask bit[1]
                | 0x1 << 6  // vt_track_err --> status bit[6], mask bit[2]
                | 0x1 << 7  // watchdog_err --> status bit[7], mask bit[3]
                ,
  .FatalMaskCh1 = 0x0,   // Not channelized
  
  .LockOffset   = 0x00c,
  .IsrOffset    = 0x100, // Status  -- wtc
  .ItrOffset    = 0x104, // Trigger -- wo
  .ImrOffset    = 0x114, // Mask rd -- ro
  .IerOffset    = 0x118  // Enable  -- wo
};



static IntDataTable MCIntData = {
  .IRQ          = 179U,
  .BaseAddr     = { 0xF6150000, 0xF62C0000, 0xF6430000, 0xF65A0000 },
  .CEMaskCh0    =  0x3 << 15 , // CE[0,1]_ECC: Correctable ECC CH0
  .CEMaskCh1    =  0x3 << 17 , // CE[0,1]_ECC: Correctable ECC CH1
  .FatalMaskCh0 =  0xF         // NSU_FATAL[3:0]
                |  0x3 <<  9   // UC[0,1]_ECC: Uncorrectable ECC CH0
                |  0x1 << 19   // NA_CMD_FATAL: QOS Q error
                |  0x1 << 21   // DC_CMD_FATAL: Ch0 cmd Q error
                |  0x1 << 27   // DRAM_PARITY_FATAL_0 cmd/addr parity
                , 
  .FatalMaskCh1 =  0x3 << 13   // UC[0,1]_ECC: Uncorrectable ECC CH1
                |  0x1 << 23   // DC_CMD_FATAL: Ch1 cmd Q error
                |  0x1 << 28   // DRAM_PARITY_FATAL_1 cmd/addr parity
                ,
  .LockOffset   = 0x0c, // Lock          
  .IsrOffset    = 0x14, // Status  -- wtc
  .ItrOffset    = 0x18, // Trigger -- wo 
  .ImrOffset    = 0x1C, // Mask rd -- ro 
  .IerOffset    = 0x20  // Enable  -- wo 
};

static XScuGic InterruptController;  /* Interrupt controller instance */



/*****************************************************************************/
/**
 * Interrupt hanlder iterates through each enabled Memory Controller and 
 * determines if: 
 *   -- any fatal errors are set
 *   -- any correctable ECC errors occurred.
 *
 * ECC errors status gets cleared and the EMIO output is pulsed.  
 *
 * If a fatal event is found, the EMIO is asserted until reset.
 *
 * @param	data is a pointer to the interrupt data table.
 *
 * @return void
 *
 * @note	None.
 *
 ******************************************************************************/
void REMInterruptHandler(void* data) {
  IntDataTable* RowPtr = (IntDataTable*) data; 
  u32 Fatal;
  u32 CorECC;
  u32 StatusReg;

  //TRACE_PRINT("Interrupt fired: %d\n\r", RowPtr->IRQ);

  for (int i=0; i<=REM_NUM_CONTROLLERS; i++) {
    if (0x1 & (REM_ENABLED_MC_MAP >> i)) { // Ignore unused controllers
      StatusReg = Xil_In32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset);
      Fatal  = StatusReg & RowPtr->FatalMaskCh0;
      CorECC = StatusReg & RowPtr->CEMaskCh0;
      
      // Only check against Channel 1 mask when it's enabled on that MC
      if ((REM_DUAL_CHANNEL_MAP >> i) & 0x1) {
        Fatal  |= StatusReg & RowPtr->FatalMaskCh1;
        CorECC |= StatusReg & RowPtr->CEMaskCh1;
      }


      if (Fatal) {
        TRACE_PRINT("IRQ:%d   Bank:%d  Status --> Fatal :%08x\n\r",RowPtr->IRQ,i,Fatal);
        // Assert EMIO Fatal bit
        wrGPIO(REM_GPIO_FATAL_MASK, REM_GPIO_FATAL_MASK);

        // Unlock Register Access & clear ISR. PCSR access is only good for a single
        // transaction and then it re-locks
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, REM_PCSR_KEY);
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset, Fatal);
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, 0x1);
      }

      if (CorECC) {
        TRACE_PRINT("IRQ:%d   Bank:%d  Status --> CorECC:%08x\n\r",RowPtr->IRQ,i,CorECC);
        // Assert EMIO Correctable Error bit
        wrGPIO(REM_GPIO_CECC_MASK, REM_GPIO_CECC_MASK);

        // Unlock Register Access & clear ISR
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, REM_PCSR_KEY);
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset, CorECC);
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, 0x1);
      
        // Clear EMIO Correctable Error bit
        wrGPIO(REM_GPIO_CECC_MASK, 0x0);
      }
    }
  }

  // Re-enable interrupt
  XScuGic_Enable(&InterruptController, RowPtr->IRQ);
}

/****************************************************************************/
/**
*
* This function setups the GIC interrupt system.
*
* @param	IntcDeviceID is the unique ID of the interrupt controller
* @param	IntcInstacePtr is a pointer to the interrupt controller
*		instance.
*
* @return	REM_SUCCESS if successful, otherwise REM_FAILURE.
*
* @note		None.
*
*****************************************************************************/
static int SetupInterruptSystem(u16 IntcDeviceID,
				    XScuGic* IntcInstancePtr)
{
	int Status;
	XScuGic_Config* IntcConfig; /* The configuration parameters of the
					interrupt controller */

	/*
	 * Initialize the interrupt controller driver
	 */
	IntcConfig = XScuGic_LookupConfig(IntcDeviceID);
	if (NULL == IntcConfig) {
		return REM_FAILURE;
	}

	Status = XScuGic_CfgInitialize(IntcInstancePtr, IntcConfig,
					IntcConfig->CpuBaseAddress);
	if (Status != REM_SUCCESS) {
		return REM_FAILURE;
	}

	/*
	 * Connect the interrupt controller interrupt handler to the hardware
	 * interrupt handling logic in the ARM processor.
	 */
	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_INT,
			(Xil_ExceptionHandler) XScuGic_InterruptHandler,
			IntcInstancePtr);

	/*
	 * Enable interrupts in the ARM
	 */
	Xil_ExceptionEnable();

	return REM_SUCCESS;
}


/****************************************************************************/
/**
*
* Iterate through each enabled Memory Controller and validate the initial 
* 'Config Done' boot-up status completed without error.
*
* @param	map -- bit-map indicating which memory controllers are active
*
* @return	REM_SUCCESS when all enabled controllers reach CALDONE state without 
*                     any errors or timing out.
*   else: REM_FAILURE
*
* @note		None.
*
*****************************************************************************/
static int CheckDDRInitStatus()
{
  u32 rslt;

  for (int i=0; i<=REM_NUM_CONTROLLERS; i++) {

    if ((0x1 << i) & REM_ENABLED_MC_MAP) { // Ignore unused controllers

      /*
       * Wait for CALDONE 
       *   The appropriate time is highly appication dependant and depending on: 
       *     -- ECC fill is enabled
       *     -- memory speed 
       *     -- memory size
       *  Upper bound could be over a full minute wait in extreme cases
       */

      for (int timeout = REM_TIMEOUT_WAIT; timeout > 0; timeout--) {
        rslt = Xil_In32(UBIntData.BaseAddr[i] + REM_UB_PCSR_STATUS_OFFSET);
        if (0x10 == (rslt & 0x38) ) { 
          // Validate (~CALERROR:5 & CALDONE:4 & ~INCAL:3) --> Success! 
          TRACE_PRINT("CheckDDRInitStatus .... Good: Controller:%d  timeout=%d\r\n",
                       i,timeout);
          break;

        } else if (rslt & 0x20) { 
          // Catch CALERROR:5 set  
          TRACE_PRINT("CheckDDRInitStatus .... error: CALERROR asserted\r\n");
          return REM_FAILURE;

        } else {
          // The typical wait case is:  (rslt & 0x8) --> is INCAL:3 still active?  
          // However, all other failure-to-thrive cases can be caught 
          // with a generic general timeout
          if (timeout == 0) {
            TRACE_PRINT("CheckDDRInitStatus .... error: Controller %d timed out\r\n",i);
            TRACE_PRINT("     UB_PCSR_SATUS = %x \r\n",rslt);
            return REM_FAILURE;
          }
          usleep(1000);
        }
      }
    }
  }

  // Failure cases are explicitly covered --> implicit success 
  return REM_SUCCESS;
}


/****************************************************************************/
/**
*
* Write select bits of the 16-bit GPIO register.  Upper 16-bits are the mask
* and lower 16-bits are the data value applied for all zeroed mask bits.
*
* @param	enable -- bit-map indicating which bits to update
* @param	data   -- value of data to write to the enabled GPIO bits
*
* @return	void
*
* @note		None.
*
*****************************************************************************/
static void wrGPIO(u16 enable, u16 data) {
  Xil_Out32(REM_MASK_DATA_3_MSW, (~((u32) enable)) << 16 | (u32) data);
}
  

/****************************************************************************/
/**
*
* For the given interrupt data, enable unmask the bits of interest in the 
* the interrupt enable register.
*
* @param	RowPtr -- pointer to interrupt info data struct
*
* @return	void
*
* @note		None.
*
*****************************************************************************/
static void UnmaskMCBits(IntDataTable* RowPtr) {
  u32 UnmaskIMR;

  for (int i=0; i<=REM_NUM_CONTROLLERS; i++) {
    if (0x1 & (REM_ENABLED_MC_MAP >> i)) { // Skip unused controllers
      UnmaskIMR = RowPtr->FatalMaskCh0 |  RowPtr->CEMaskCh0;
      
      // Only check against Channel 1 mask when it's enabled on that MC
      if ((REM_DUAL_CHANNEL_MAP >> i) & 0x1) {
        UnmaskIMR |= RowPtr->FatalMaskCh1;
        UnmaskIMR |= RowPtr->CEMaskCh1;
      }

      if (Xil_In32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset) & UnmaskIMR) {
        TRACE_PRINT("WARNING: IRQ %d, MC %d already has ISR bits set: Enable:%08x vs ISR %08x\n\r", UnmaskIMR, Xil_In32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset));
        
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, REM_PCSR_KEY);
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->IsrOffset,  UnmaskIMR  );
        Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, 0x1);
      }

#ifdef DO_TRACE
      {
        u32 Mask;

        Mask = Xil_In32(RowPtr->BaseAddr[i] + RowPtr->ImrOffset);
        TRACE_PRINT("IRQ %d, MC %d  Mask pre :%08x \n\r",RowPtr->IRQ, i, Mask);
  
        if (~Mask & UnmaskIMR) { 
          TRACE_PRINT("WARNING: IRQ %d, MC %d already has IMR bits cleared: Mask:%08x vs %08x\n\r",
              RowPtr->IRQ, i, Mask, UnmaskIMR);
        }
      }
#endif

      Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, REM_PCSR_KEY);
      Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->IerOffset,  UnmaskIMR  );
      Xil_Out32(RowPtr->BaseAddr[i] + RowPtr->LockOffset, 0x1);

#ifdef DO_TRACE
      {
        u32 Mask;

        Mask = Xil_In32(RowPtr->BaseAddr[i] + RowPtr->ImrOffset);
        TRACE_PRINT("IRQ %d, MC %d  Mask post:%08x \n\r",RowPtr->IRQ, i, Mask);
  
        if (Mask & UnmaskIMR) { 
          TRACE_PRINT("ERROR: IRQ %d, MC %d IMR bits not cleared: Mask:%08x vs %08x\n\r",
             RowPtr->IRQ, i, Mask, UnmaskIMR);
       }
     }
#endif

    }
  }
}

/*****************************************************************************/
/**
*
* Main
*
* @param	None
*
* @return
*  Never returns
*
* @note		None
*
*****************************************************************************/
int main(void)
{
  WAIT_FOR_PLM_DONE();

  /*
   * DataTable array is populated first with interrupts, secondly with
   * poll-based registers.  
   *
   * This crudely allows the interrupts to be setup differently from 
   * the poll-based registers while using the same data struct
   * 
   * Thus, 0 to (REM_NUM_INTS -1) is generically used for full-interrupt handling
   * mechanisms, and REM_NUM_INTS to end is polling.
   */
  IntDataTable DataTable[REM_DATA_TABLE_SZ] = {MCIntData,UBIntData};
  IntDataTable* DataRow;
  int Status;
  u32 rmw;

#ifdef DO_TRACE
  /*
   * Wait for PDI done from PLM in order get a clean console output
   */
  while(0 == (*(volatile uint32_t*)0xF1110000 & 0x10)) {
    ;
  }
#endif
  TRACE_PRINT("Build info: %s %s\n\r",__DATE__,__TIME__);


  /* 
   * Configure EMIO to drive as outputs
   *
   * R5-0 also utilizes the EMIO -- These configuration registers could
   * potentially collided with the other processor's config, so we mutex 
   * the read-mod-write operation.
   *
   * Since the set/clear operations are broken up into different 16-bit 
   * sub-registers, there's no issue when assigning the output values later.
   */
  r5GrabSysMutex();
  rmw = Xil_In32(REM_DIRM_3);
  Xil_Out32(REM_DIRM_3, (REM_GPIO_CFGDONE_MASK | REM_GPIO_CECC_MASK | REM_GPIO_FATAL_MASK)<<16 | rmw);
  rmw = Xil_In32(REM_OEN_3);
  Xil_Out32(REM_OEN_3,  (REM_GPIO_CFGDONE_MASK | REM_GPIO_CECC_MASK | REM_GPIO_FATAL_MASK)<<16 | rmw);
  r5ReleaseSysMutex();

	/*
	 * Validate the memory controllers completed setup error free "CALDONE"
   * PLM actually takes care of this process prior to releasing PL to run.
   * Thus, since this design currently signals the PL upon an error, it
   * this routine is actually redundant.
   * That said, we'll leave the check here as belt+suspenders as a reference
   * for other use cases should the signaling be destined somehwere other than
   * PL.
	 */
  Status = CheckDDRInitStatus();
	if (Status == REM_SUCCESS) {
    wrGPIO(REM_GPIO_CFGDONE_MASK,REM_GPIO_CFGDONE_MASK);
	} else {
    wrGPIO(REM_GPIO_FATAL_MASK,REM_GPIO_FATAL_MASK);
  }

	/*
	 * Connect the Intc to the interrupt subsystem such that interrupts can
	 * occur. This function is application specific.
	 */
	Status = SetupInterruptSystem(XPAR_SCUGIC_SINGLE_DEVICE_ID, &InterruptController);
	if (Status != REM_SUCCESS) {
    TRACE_PRINT("SetupInterruptSystem  ... Error \n\r");
	} 

	/*
	 * Connect REMInterruptHandler to the interrupt controller 
	 */
  for (int i=0; i<REM_NUM_INTS; i+=1) {
	  DataRow = &DataTable[i];
	  Status = XScuGic_Connect(&InterruptController, DataTable[i].IRQ,
	  	                        REMInterruptHandler, DataRow);
	  if (Status != REM_SUCCESS) {
      TRACE_PRINT("XScuGic_Connect ... Error \n\r");
		  return REM_FAILURE;
	  }
  }

  /*
   * Unmask interrupts at MC's
   */
  for (int i=0; i<REM_NUM_INTS; i+=1) {
    UnmaskMCBits(&DataTable[i]);
  }
  TRACE_PRINT("Unmasked.\n\r");

  /*
   * Finally, enable each Interrupt at GIC
   */
  for (int i=0; i<REM_NUM_INTS; i+=1) {
	  XScuGic_Enable(&InterruptController, DataTable[i].IRQ);
  }
  TRACE_PRINT("Interrupts enabled \n\r");


#ifdef DO_TEST
  // Test code
  
  TRACE_PRINT("Starting trigger walk ... for MemController IRQ: %0d\n\r",DataTable[0].IRQ);

  for (int i = 0; i<32; i++) {
       
    TRACE_PRINT("%0d --> Bit %2d -- Expect hit:%d \n\r",DataTable[0].IRQ,i, 0x1 & (~0xE75019F0 >> i) );

    // unlock PCSR
    Xil_Out32(DataTable[0].BaseAddr[0] + DataTable[0].LockOffset, REM_PCSR_KEY);
    Xil_Out32(DataTable[0].BaseAddr[0] + DataTable[0].ItrOffset, 0x1<<i);
    Xil_Out32(DataTable[0].BaseAddr[0] + DataTable[0].LockOffset, 0x1);

	  //XScuGic_Enable(&InterruptController, DataTable[1].IRQ);
   
    usleep(1000);
  }
#endif // DO_TEST

#ifdef DO_TRACE
  for (int i = 0; i<REM_NUM_CONTROLLERS; i++) {
    if (0x1 & (REM_ENABLED_MC_MAP >> i)) { // Skip unused controllers
      TRACE_PRINT("Example XSCT writes to fake an ECC correctable on bank %0d:\n\r mwr 0x%08x 0x%08x \r\n mwr 0x%08x 0x%08x\n\r", i,
        DataTable[0].BaseAddr[i] + DataTable[0].LockOffset, REM_PCSR_KEY,
        DataTable[0].BaseAddr[i] + DataTable[0].ItrOffset, DataTable[0].CEMaskCh0);


      TRACE_PRINT("Example XSCT writes to fake a UB fatal on Bank %0d:\n\r mwr 0x%08x 0x%08x \r\n mwr 0x%08x 0x%08x\n\r", i,
        DataTable[1].BaseAddr[i] + DataTable[1].LockOffset, REM_PCSR_KEY,
        DataTable[1].BaseAddr[i] + DataTable[1].ItrOffset, DataTable[1].FatalMaskCh0);
    }
  }
#endif // DO_TRACE
  
  TRACE_PRINT("Setup complete ... entering poll loop \r\n");

  // Poll the status registers that RPU/APU is unable to get an interrupt from
  while (1) {
    for (int i=REM_NUM_INTS; i < REM_DATA_TABLE_SZ; i++) {
      REMInterruptHandler(&DataTable[i]);
    }
    usleep(1000);
  }


  __builtin_unreachable();
  return 0;
}
