/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

// FIXME - apply timeouts

static void Pause(Sys* pSys,uint32_t* pqTimeoutMs)
{
    if(*pqTimeoutMs > 0) {
        *pqTimeoutMs -= 1;
        usleep(1000);
    }
}

bool WaitForSyncWord(Sys* pSys,uint32_t qTimeoutMs)
{
    uint32_t qSync;

//    Verbose(pSys,"%s()",__func__);
    for(;;) {
        qSync = le32toh(pSys->pMA->tResponseArea.qFEEDC0DE);
//        Verbose(pSys,"\tGot %u",qSync);
        if(FEEDC0DE == qSync) {
            return true;
        }
        if(qTimeoutMs > 0) {
            Pause(pSys,&qTimeoutMs);
        } else {
            return false;
        }
    }
    /*
     * Not reached
     */
    return false;
}

static uint32_t IssueCommandAndWait(Sys* pSys,MessageArea* pMA,uint32_t qTimeoutMs)
{
    uint32_t qRet;
    char pcCommandString[COMMAND_STRING_BUFSIZE];
    volatile uint32_t qCrc32;

    qCrc32 = Crc32_ComputeBuf(
                0,                      /* uint32_t inCrc32, */
                &pMA->tCommandArea,     /* const void *buf, */
                sizeof(CommandArea) - 4 /* uint32_t bufLen */
            );

    /*
     * NOTE:
     * The live message area will be acted on if the checksum is written.
     * The verbose print may look at it after the r5 has acted on and wiped it out
     * Thus, post the checksum after the printf
     */
    Verbose(pSys,"%s(): %s %08x %08x %08x %08x %08x %08x %08x %08x ... %08x",
            __func__,
            GetCommandString(pMA->tCommandArea.qCommand,pcCommandString),
            pMA->tCommandArea.pqArgs[0],
            pMA->tCommandArea.pqArgs[1],
            pMA->tCommandArea.pqArgs[2],
            pMA->tCommandArea.pqArgs[3],
            pMA->tCommandArea.pqArgs[4],
            pMA->tCommandArea.pqArgs[5],
            pMA->tCommandArea.pqArgs[6],
            pMA->tCommandArea.pqArgs[7],
            qCrc32
          );
    qCrc32 = htole32(qCrc32);
    pMA->tCommandArea.qChecksum = qCrc32;

    for(;;) {
        if(le32toh(pMA->tResponseArea.qChecksum) == Crc32_ComputeBuf(
                                                0,                          /* uint32_t inCrc32, */
                                                &pMA->tResponseArea,        /* const void *buf, */
                                                sizeof(ResponseArea) - 8    /* uint32_t bufLen */
                                            )
          ) {
            pMA->tResponseArea.qChecksum ^= ~0;
            qRet = le32toh(pMA->tResponseArea.qErrCode);
            break;
        } else if(qTimeoutMs > 0) {
            Pause(pSys,&qTimeoutMs);
        } else {
            qRet = ERR_TIMEOUT;
            break;
        }
    }

    Verbose(pSys,"\t-> %u",qRet);
    return qRet;
}

uint32_t VerifyFlash(
    Sys* pSys,
    uint32_t qConnectionMode,
    uint32_t qBusWidth,
    uint32_t qQspiPeripheralInputSpeedHz,
    uint32_t qPrescaler,
    uint32_t qCacheCoherentDMA,
    uint32_t qPEBSize,
    uint32_t qNumPEBs,
    uint32_t qSubsectorSize,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    int iArg;
    uint32_t qRet;

    Verbose(pSys,"%s(%u,%u,%u,%u,%u,%u,%u,%u,%p)",
            __func__,
            qConnectionMode,
            qBusWidth,
            qQspiPeripheralInputSpeedHz,
            qPrescaler,
            qCacheCoherentDMA,
            qPEBSize,
            qNumPEBs,
            qSubsectorSize,
            (void*)pqErrBSP
            );
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_VERIFY_FLASH);
    iArg = 0;
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qConnectionMode);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qBusWidth);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qQspiPeripheralInputSpeedHz);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qPrescaler);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qCacheCoherentDMA);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qPEBSize);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qNumPEBs);
    pMA->tCommandArea.pqArgs[iArg++] = htole32(qSubsectorSize);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_VERIFY_FLASH);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t VerifyImage(
    Sys* pSys,
    uint32_t qImageOffset,
    uint32_t qImageSize,
    uint32_t qImageChecksum,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u,%u,%u)",__func__,qImageOffset,qImageSize,qImageChecksum);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_VERIFY_IMAGE);
    pMA->tCommandArea.pqArgs[0] = htole32(qImageOffset);
    pMA->tCommandArea.pqArgs[1] = htole32(qImageSize);
    pMA->tCommandArea.pqArgs[2] = htole32(qImageChecksum);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_VERIFY_IMAGE);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t GetMultibootBytes(
    Sys* pSys,
    OUTPUT_PARAMETER uint32_t* pqErrBSP,
    OUTPUT_PARAMETER uint32_t* pqMultibootBytes
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%p)",__func__,(void*)pqMultibootBytes);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_GET_MULTIBOOT_BYTES);
    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_GET_MULTIBOOT_BYTES);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    *pqMultibootBytes = le32toh(pMA->tResponseArea.pqData[0]);

    return qRet;
}

uint32_t ErasePEBs(
    Sys* pSys,
    uint32_t qPEBStart,
    uint32_t qPEBEnd,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u,%u)",__func__,qPEBStart,qPEBEnd);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_ERASE_PEBS);
    pMA->tCommandArea.pqArgs[0] = htole32(qPEBStart);
    pMA->tCommandArea.pqArgs[1] = htole32(qPEBEnd);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_ERASE_PEB*(qPEBEnd - qPEBStart + 1));
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t ReadPEBInternal(
    Sys* pSys,
    uint32_t qPEBNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u)",__func__,qPEBNumber);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_READ_PEB);
    pMA->tCommandArea.pqArgs[0] = htole32(qPEBNumber);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_READ_PEB);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t ReadPEB(
    Sys* pSys,
    uint32_t qPEBNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP,
    OUTPUT_PARAMETER uint8_t* pbDst
    )
{
    uint32_t qRet;

    Verbose(pSys,"%s(%u,%p)",__func__,qPEBNumber,pbDst);
    qRet = ReadPEBInternal(pSys,qPEBNumber,pqErrBSP);

    if(ERR_NONE == qRet) {
        memcpy(pbDst,pSys->pAA,pSys->pBI->qPEBSize);
    }

    return qRet;
}

uint32_t WritePEBInternal(
    Sys* pSys,
    uint32_t qPEBNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u)",__func__,qPEBNumber);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_WRITE_PEB);
    pMA->tCommandArea.pqArgs[0] = htole32(qPEBNumber);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_WRITE_PEB);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t WritePEB(
    Sys* pSys,
    uint32_t qPEBNumber,
    const uint8_t* pbSrc,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    Verbose(pSys,"%s(%u,%p)",__func__,qPEBNumber,pbSrc);
    memcpy(pSys->pAA,pbSrc,pSys->pBI->qPEBSize);

    return WritePEBInternal(pSys,qPEBNumber,pqErrBSP);
}

uint32_t EraseSubsector(
    Sys* pSys,
    uint32_t qSubsectorNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u)",__func__,qSubsectorNumber);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_ERASE_SUBSECTOR);
    pMA->tCommandArea.pqArgs[0] = htole32(qSubsectorNumber);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_ERASE_SUBSECTOR);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t ReadSubsectorInternal(
    Sys* pSys,
    uint32_t qSubsectorNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u)",__func__,qSubsectorNumber);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_READ_SUBSECTOR);
    pMA->tCommandArea.pqArgs[0] = htole32(qSubsectorNumber);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_READ_SUBSECTOR);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t ReadSubsector(
    Sys* pSys,
    uint32_t qSubsectorNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP,
    OUTPUT_PARAMETER uint8_t* pbDst
    )
{
    uint32_t qRet;

    Verbose(pSys,"%s(%u,%p)",__func__,qSubsectorNumber,pbDst);
    qRet = ReadSubsectorInternal(pSys,qSubsectorNumber,pqErrBSP);
    if(ERR_NONE == (ERR)qRet) {
        memcpy(pbDst,pSys->pAA,pSys->pBI->qSubsectorSize);
    }

    return qRet;
}

uint32_t WriteSubsectorInternal(
    Sys* pSys,
    uint32_t qSubsectorNumber,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s(%u)",__func__,qSubsectorNumber);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_WRITE_SUBSECTOR);
    pMA->tCommandArea.pqArgs[0] = htole32(qSubsectorNumber);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_WRITE_SUBSECTOR);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}

uint32_t WriteSubsector(
    Sys* pSys,
    uint32_t qSubsectorNumber,
    const uint8_t* pbSrc,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    Verbose(pSys,"%s(%u,%p)",__func__,qSubsectorNumber,pbSrc);
    memcpy(pSys->pAA,pbSrc,pSys->pBI->qSubsectorSize);
    return WriteSubsectorInternal(pSys,qSubsectorNumber,pqErrBSP);
}

uint32_t SimStop(
    Sys* pSys,
    OUTPUT_PARAMETER uint32_t* pqErrBSP
    )
{
    MessageArea* pMA;
    uint32_t qRet;

    Verbose(pSys,"%s()",__func__);
    pMA = (MessageArea*)pSys->pMA;
    pMA->tCommandArea.qCommand = htole32(COMMAND_SIM_STOP);

    qRet = IssueCommandAndWait(pSys,pMA,TIMEOUT_SIM_STOP);
    *pqErrBSP = le32toh(pMA->tResponseArea.qErrBSP);
    return qRet;
}
